import numpy as np
from .primitive import PointCloud as PointCloud
from typing import Tuple

class Corners(PointCloud):
    def get(self) -> np.ndarray: ...
    @staticmethod
    def product(shape: Tuple) -> Corners: ...



from itertools import product
from typing import Tuple

import numpy as np

from .primitive import PointCloud


class Corners(PointCloud):
    def get(self) -> np.ndarray:
        corners = super().get()
        n, p = corners.shape
        if 2 ** p != n:
            msg = ""
            raise ValueError(msg)
        return corners

    @staticmethod
    def product(shape: Tuple) -> "Corners":
        masks = list(product((False, True), repeat=len(shape)))
        corners = [[s if c else 0 for c, s in zip(mask, shape)] for mask in masks]
        return Corners(corners)



import typing as tp

import numpy as np


def rotate(
    pivot_point: tp.Iterable[float],
    turning_point: tp.Iterable[float],
    angle: tp.Iterable[float],
) -> np.ndarray:
    pivot_point = np.asarray(pivot_point)
    turning_point = np.asarray(turning_point)
    assert pivot_point.shape == turning_point.shape
    rotate_matrix = get_rotate_matrix(pivot_point.shape[-1], angle)

    turning_point = turning_point - pivot_point
    turning_point = rotate_matrix @ turning_point
    turning_point += np.asarray(pivot_point)
    return turning_point


def get_rotate_matrix(dim: int, angle: tp.Iterable[float]) -> np.ndarray:
    if dim == 2:
        return get_rotate_matrix_2d(angle[0])
    elif dim == 3:
        return get_rotate_matrix_3d(angle)
    else:
        raise ValueError(f"ndim should be 2 or 3, not {dim}")


def get_rotate_matrix_2d(angle: float) -> np.ndarray:
    return np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])


def get_rotate_matrix_3d(angle: tp.Iterable[float]) -> np.ndarray:
    rotate_matrix_x = np.array(
        [
            [1, 0, 0],
            [0, np.cos(angle[0]), -np.sin(angle[0])],
            [0, np.sin(angle[0]), np.cos(angle[0])],
        ]
    )
    rotate_matrix_y = np.array(
        [
            [np.cos(angle[1]), 0, np.sin(angle[1])],
            [0, 1, 0],
            [-np.sin(angle[1]), 0, np.cos(angle[1])],
        ]
    )
    rotate_matrix_z = np.array(
        [
            [np.cos(angle[2]), -np.sin(angle[2]), 0],
            [np.sin(angle[2]), np.cos(angle[2]), 0],
            [0, 0, 1],
        ]
    )
    return rotate_matrix_x @ rotate_matrix_y @ rotate_matrix_z



import numpy as np
import typing as tp

def rotate(pivot_point: tp.Iterable[float], turning_point: tp.Iterable[float], angle: tp.Iterable[float]) -> np.ndarray: ...
def get_rotate_matrix(dim: int, angle: tp.Iterable[float]) -> np.ndarray: ...
def get_rotate_matrix_2d(angle: float) -> np.ndarray: ...
def get_rotate_matrix_3d(angle: tp.Iterable[float]) -> np.ndarray: ...



from pytest import mark
from pyece.corners import Corners

@mark.parametrize(
    "shape",
    [
        (1,2),
        (1,2,3),
        (1,2,3,4)
    ],
)
def test_corners_product(shape):
    cs = Corners.product(shape)
    assert cs.dim == len(shape)
    assert cs.centre.mean() == sum(shape)/len(shape)/2



from abc import ABC, abstractmethod
from typing import Any, Callable, Iterable, Sequence, Union

import numpy as np

from .math.rotate import rotate

LikeProperty = Union["Property", float, int, str]
PropertySequence = Sequence[LikeProperty]


class Property(ABC):
    @abstractmethod
    def get(self) -> Any:
        return NotImplemented

    @property
    def value(self) -> Any:
        return self.get()

    def transform(self, operation: "Operation", **kwargs) -> "Property":
        assert isinstance(operation, Operation)
        return operation(**kwargs)(self)


class ConstantProperty(Property):
    def __init__(self, value):
        self._value = value

    def get(self) -> Any:
        return self._value


def as_property(value: LikeProperty) -> Property:
    if isinstance(value, Property):
        return value
    return ConstantProperty(value)


class Source(Property):
    def __init__(self, fn: Callable, *args, **kwargs):
        assert callable(fn)
        self._fn = fn
        self._args = args
        self._kwargs = kwargs

    def get(self) -> Any:
        return self._fn(*self._args, **self._kwargs)


class Convert(Source):
    def __init__(self, p: LikeProperty, fn: Callable, *args, **kwargs):
        self._original = as_property(p)
        super().__init__(fn, *args, **kwargs)

    def get(self) -> Any:
        return self._fn(self._original.value, *self._args, **self._kwargs)


class Iter(Property):
    def __init__(self, seq: Iterable):
        self._source = seq
        self._iter = iter(self._source)

    def _next(self):
        try:
            return next(self._iter)
        except StopIteration:
            self._iter = iter(self._source)
            return next(self._iter)

    def get(self) -> Any:
        return self._next()


class RandomUniform(Property):
    def __init__(self, low=0.0, high=1.0):
        self._low = low
        self._high = high

    def get(self):
        return np.random.uniform(self._low, self._high)


class Point(Property):
    def __init__(self, point: PropertySequence):
        self._point = [as_property(p) for p in point]

    def get(self) -> np.ndarray:
        return np.asarray([p.value for p in self._point])


LikePoint = Union[Point, PropertySequence]


def as_point(value: LikePoint) -> Point:
    if isinstance(value, Point):
        return value
    return Point(value)


class PointCloud(Point):
    def __init__(self, points: Sequence[LikePoint]):
        super().__init__([as_point(p) for p in points])

    def transform(self, operation: "Operation", **kwargs) -> "Point":
        if isinstance(operation, PointOperation):
            points = self.value
            if isinstance(operation, (PointRotate, PointInflation)):
                if operation._pivot is None:
                    kwargs["pivot"] = points.mean(0)
            fn = operation(**kwargs)
            return PointCloud(map(fn, points))
        return super().transform(operation, **kwargs)


LikePointCloud = Union[PointCloud, Sequence[LikePoint]]


def as_pointcloud(value: LikePointCloud) -> PointCloud:
    if isinstance(value, PointCloud):
        return value
    return PointCloud(value)


class Operation:
    def __call__(self, **params) -> Callable:
        def inner(obj):
            self.operation(obj, **params)

        return inner

    @abstractmethod
    def operation(self, obj, **params) -> Any:
        return NotImplemented


class PointOperation(Operation):
    def __call__(self, **params) -> Callable[[LikePoint], Point]:
        def inner(point: LikePoint):
            point = as_point(point).value
            value = self.operation(point, **params)
            return Point(value)

        return inner

    @abstractmethod
    def operation(self, obj: np.ndarray, **params) -> np.ndarray:
        return NotImplemented


class PointShift(PointOperation):
    def __init__(self, shift: LikePoint):
        self._shift = as_point(shift)

    def __call__(self) -> Callable[[LikePoint], Point]:
        shift = self._shift.value
        return super().__call__(shift=shift)

    def operation(self, obj: np.ndarray, shift: np.ndarray) -> np.ndarray:
        return obj + shift


class PointRotate(PointOperation):
    def __init__(self, angle: LikeProperty, pivot: LikePoint = None):
        self._angle = as_property(angle)
        self._pivot = None if pivot is None else as_point(pivot)

    def __call__(self, pivot: LikePoint = None) -> Callable[[LikePoint], Point]:
        angle = np.asarray(self._angle.value).reshape(-1) % (2 * np.pi)
        if pivot is not None:
            pivot = as_point(pivot).value
        else:
            pivot = self._pivot.value
        return super().__call__(pivot=pivot, angle=angle)

    def operation(
        self, obj: np.ndarray, pivot: np.ndarray, angle: np.ndarray
    ) -> np.ndarray:
        return rotate(pivot, obj, angle)


class PointInflation(PointOperation):
    def __init__(self, factor: LikeProperty, pivot: LikePoint = None):
        self._factor = as_property(factor)
        self._pivot = None if pivot is None else as_point(pivot)

    def __call__(self, pivot: LikePoint = None) -> Callable[[LikePoint], Point]:
        factor = self._factor.value
        if pivot is not None:
            pivot = as_point(pivot).value
        else:
            pivot = self._pivot.value
        return super().__call__(pivot=pivot, factor=factor)

    def operation(
        self, obj: np.ndarray, pivot: np.ndarray, factor: np.ndarray
    ) -> np.ndarray:
        return pivot + (obj - pivot) * factor


class Transformator:
    def __init__(self, *operations: Operation):
        assert all([isinstance(op, Operation) for op in operations])
        self._operations = operations

    def __call__(self, instance: Property) -> Property:
        for op in self._operations:
            instance = instance.transform(op)
        return instance



import abc
import numpy as np
from .math.rotate import rotate as rotate
from abc import ABC, abstractmethod
from typing import Any, Callable, Iterable, Sequence, Union

LikeProperty: Any
PropertySequence = Sequence[LikeProperty]

class Property(ABC, metaclass=abc.ABCMeta):
    @abstractmethod
    def get(self) -> Any: ...
    @property
    def value(self) -> Any: ...
    def transform(self, operation: Operation, **kwargs) -> Property: ...

class ConstantProperty(Property):
    def __init__(self, value) -> None: ...
    def get(self) -> Any: ...

def as_property(value: LikeProperty) -> Property: ...

class Source(Property):
    def __init__(self, fn: Callable, *args, **kwargs) -> None: ...
    def get(self) -> Any: ...

class Convert(Source):
    def __init__(self, p: LikeProperty, fn: Callable, *args, **kwargs) -> None: ...
    def get(self) -> Any: ...

class Iter(Property):
    def __init__(self, seq: Iterable) -> None: ...
    def get(self) -> Any: ...

class RandomUniform(Property):
    def __init__(self, low: float = ..., high: float = ...) -> None: ...
    def get(self): ...

class Point(Property):
    def __init__(self, point: PropertySequence) -> None: ...
    def get(self) -> np.ndarray: ...
LikePoint = Union[Point, PropertySequence]

def as_point(value: LikePoint) -> Point: ...

class PointCloud(Point):
    def __init__(self, points: Sequence[LikePoint]) -> None: ...
    def transform(self, operation: Operation, **kwargs) -> Point: ...
LikePointCloud = Union[PointCloud, Sequence[LikePoint]]

def as_pointcloud(value: LikePointCloud) -> PointCloud: ...

class Operation(metaclass=abc.ABCMeta):
    def __call__(self, **params) -> Callable: ...
    @abstractmethod
    def operation(self, obj, **params) -> Any: ...

class PointOperation(Operation, metaclass=abc.ABCMeta):
    def __call__(self, **params) -> Callable[[LikePoint], Point]: ...
    @abstractmethod
    def operation(self, obj: np.ndarray, **params) -> np.ndarray: ...

class PointShift(PointOperation):
    def __init__(self, shift: LikePoint) -> None: ...
    def __call__(self) -> Callable[[LikePoint], Point]: ...
    def operation(self, obj: np.ndarray, shift: np.ndarray) -> np.ndarray: ...

class PointRotate(PointOperation):
    def __init__(self, angle: LikeProperty, pivot: LikePoint = ...) -> None: ...
    def __call__(self, pivot: LikePoint = ...) -> Callable[[LikePoint], Point]: ...
    def operation(self, obj: np.ndarray, pivot: np.ndarray, angle: np.ndarray) -> np.ndarray: ...

class PointInflation(PointOperation):
    def __init__(self, factor: LikeProperty, pivot: LikePoint = ...) -> None: ...
    def __call__(self, pivot: LikePoint = ...) -> Callable[[LikePoint], Point]: ...
    def operation(self, obj: np.ndarray, pivot: np.ndarray, factor: np.ndarray) -> np.ndarray: ...

class Transformator:
    def __init__(self, *operations: Operation) -> None: ...
    def __call__(self, instance: Property) -> Property: ...
